/* BEGIN NOTICE

This software was written by Peter DiFalco and John Sarraille.  The
algorithm employed was inspired by the paper entitled "A Fast
Algorithm To Determine Fractal Dimensions By Box Counting", which was
written by Liebovitch and Toth, and which appeared in the journal
"Physics Letters A", volume 141, pp 386-390, (1989).

This software is not warranteed: use at your own risk.  Unlimited
permission to distribute this software is granted, so long as this
notice is included in all copies made.

August 19, 1991

END NOTICE */

/*
-----------------------------------------------------------------
	FILE:	 ffddriver.c

	PURPOSE: Reads data pairs from the input file, sorts the
		 data pairs, writes the data pairs and the binary
		 representation of their mesh out to the output file.

	CALLING SEQUENCE:
		 a.out  <input file>

	PDL:	 open input file
		 get number of data pairs from input file
		 initialize array to hold data pairs
		 fill data array with data pairs
		 close input file
		 create queues
		 store array of data pairs in Qzero
		 set end-of-queue flags
		 radix sort queues
			sort ith bit of ycoord
			set end-of-queue flags
			sort ith bit of xcoord
			set end-of-queue flags
		 initialize array to hold sorted data
		 transfer data from Qzero to sorted_array
		 transfer data from Qone to sorted_array
		 destroy queues
		 open output file
		 write sorted data to output file
			write xcoord to output file
			write ycoord to output file
			write binary mesh of xcoord and ycoord to output file
		 write box_count_array to output file
		 close output file

	OUTPUT:	 Writes the sorted data pairs and the binary representation
		 of their mesh on the same line and the box_count_array
		 values to the output file.
		 
*/


#include <stdio.h>
#include "fd.h"
#include <math.h>

main(argc,argv)
	int	argc;
	char	*argv[];
{
	FILE	*infile,		/* input file */
	    	*outfile;		/* output file */
	int	i,j,m;			/* control variable */
	ulong	numpairs,	/* number of data pairs in input file */
		maxdiam=0,	/* 2^numbits - 1 */
                **data_array,	/* initial array of data pairs */
		box_count_array[numbits+1];	/* box counter array */
	float	fd, fracdim();			/* Fractal Dimension */
        double	max, min,	/* maximum and minimum values in input file */
                buf1;           /* a buffer for inputs to stay in until they
		                    are scaled and cast to integers.  */

       /* find max and min in input file.  This function should open
	  the input file for reading, scan it to get the maximum and
	  minimum data values it contains, and then seek to the
	  beginning of the input file so that the code below can
	  proceed to read the data sequentially from the beginning. We
	  are having trouble getting that to work, so we have
	  temporarily taken the expedient of closing the file in the
	  function max_min, and opening it again below.  */
       max_min (argv[1], &max, &min /*, infile*/) ;

       /* set maxdiam=2^numbits-1 -- this is the largest value that an
          unsigned integer can have with the host compiler.
          (I checked the value obtained in this manner, and it 
           is correct -- j.s.) */
        for (m=0;m<numbits;m++)	maxdiam = maxdiam + (1<<m);

	/* open input file */
	if((infile = fopen(argv[1],"r")) == NULL)
	{
		printf("Cannot open input file, %s\n",argv[1]);
		exit(-1);
	}
	 
	/* get number of data pairs from input file */
	fscanf(infile,"%lu",&numpairs);

	/* allocate array memory to hold data pairs */
        if ( (data_array = (ulong **) malloc (2*sizeof(ulong *))) == NULL)
	{
		printf("Memory allocation failure (data_array).\n");
		exit(-1);
	}

	for (i=0;i<2;i++)
          if ((data_array[i] = (ulong *) malloc (numpairs*sizul)) == NULL)
	    {
		printf("Memory allocation failure (data_array).\n");
		exit(-1);
            }

	/* fill data array with data pairs, scaling the data as we go to be
	   integers between 0 and maxdiam (inclusive). */
	for(i=0;i<numpairs;i++)
	{
           fscanf(infile,"%lf",&buf1);
           data_array[0][i] = (ulong)( (buf1-min) * (maxdiam/(max-min)) );
           fscanf(infile,"%lf",&buf1);
           data_array[1][i] = (ulong)( (buf1-min) * (maxdiam/(max-min)) );
	}

	/* close the input file */
	fclose(infile);


	/* radix sort queues */
	radixsort(data_array,2,numpairs);

	/* sweep data_array */
	sweep(data_array,box_count_array,2,numpairs);

	/* write box counts to standard output. */
        if (give_counts == 1)
          {
            for(i=0;i<=numbits;i++)
		printf("box_count_array[%d] =\t%lu\n",i,box_count_array[i]);
          }
	/* find and print Fractal Dimension */
	fd = fracdim(box_count_array,numpairs);
	printf("\n\nFractal Dimension %f\n",fd);

	/* write sorted data to output file */
	if(outputfile == 1)
	{	/* open output file */
		if((outfile = fopen(argv[2],"w")) == NULL)
		{
			printf("Cannot open output file, %s.\n",argv[2]);
			exit(-1);
		}

		/* write sorted data to output file */
		for(i=0;i<numpairs;i++)
		{	/* write xcoord to output file */
			fprintf(outfile,"%10lu\t",data_array[0][i]);

			/* write ycoord to output file */
			fprintf(outfile,"%10lu\t",data_array[1][i]);
			
			/* write binary mesh of xcoord and ycoord to output file */
			for(j=(numbits-1);j>=0;j--)
			{
				if(IS_A_ONE(data_array[0][i],j))
					fprintf(outfile,"%d",1);
				else
					fprintf(outfile,"%d",0);
				if(IS_A_ONE(data_array[1][i],j))
					fprintf(outfile,"%d",1);
				else
					fprintf(outfile,"%d",0);
			}
			fprintf(outfile,"\n");
		}
		fprintf(outfile,"\n");
		
		/* write box_count_array to output file */
		for(i=0;i<=numbits;i++)
			fprintf(outfile,"box_count_array[%d] =\t%lu\n",i,box_count_array[i]);

		/* write Fractal Dimension to output file */
		fprintf(outfile,"\nFractal Dimension: %f\n",fd);

		/* close output file */
		if (outputfile == 1) fclose(outfile);
		printf("closed output file");
	}
}
