/* BEGIN NOTICE

This software was written by Peter DiFalco and John Sarraille.  The
algorithm employed was inspired by the paper entitled "A Fast
Algorithm To Determine Fractal Dimensions By Box Counting", which was
written by Liebovitch and Toth, and which appeared in the journal
"Physics Letters A", volume 141, pp 386-390, (1989).

This software is not warranteed: use at your own risk.  Unlimited
permission to distribute this software is granted, so long as this
notice is included in all copies made.

August 19, 1991

END NOTICE */

/*
function fdutil.c
   Contains utility functions for Fractal Dimension Programs

radixsort(array,d1,d2) -- Uses two queues and the queue package to radix
	sort the "array" with dimensions d1 X d2. The "array" is passed
	by reference, so no return is necessary.
sweep(array,d1,d2,count_array) -- Performs a box count sweep of "array"
	with dimensions d1 X d2 and stores the box count in "count_array".
	The "count_array" is passed by reference, so no return is necessary.
	The actual sweep is performed by storing the result of an XOR on two
	x values in xtest and the result of an XOR on two y values in ytest,
	checking each bit of the mesh of xtest and ytest for a one, and if
	the current kth bit of the mesh is a one, then increment each
	count_array[m] where m<=k.
*/

#include <stdio.h>
#include "fd.h"
#include <math.h>

void max_min();
void radixsort();
void sweep();
float fracdim();

void max_min(userinputfile,pmax, pmin /*, infile*/)
	char *userinputfile;
	double *pmax, *pmin ;
	/* FILE	*infile;*/	/* input file */
	
{
	FILE	*infile;	/* input file */
	double	temp;
	int	i;		/* control variable */
	ulong	numpairs;	/* number of data pairs in input file */

	/* open input file */
	if((infile = fopen(userinputfile,"r")) == NULL)
	{
		printf("Cannot open input file, %s\n",userinputfile);
		exit(-1);
	}

	/* get numpairs */
	if ((fscanf(infile,"%lu",&numpairs)) == 0)
	  {
	     printf ("Format of input file, %s, ", userinputfile);
	     printf ("is incorrect -- please check.\n");
	     exit(-1);
	  }

	/* find maximum and minimum data values in the input file.
	   These are needed so that the data can be scaled to be a set of
	   non-negative integers between 0 and maxdiam, where maxdiam will
	   be the largest integer value expressible as an element of the
	   data type "unsigned long int". */
	fscanf(infile,"%lf",&temp);
	*pmin=*pmax=temp;
	for(i=0;i<(2*numpairs-1);i++)
	{
		fscanf(infile,"%lf",&temp);
		if(temp > *pmax)
			*pmax = temp;
		else
			if(temp < *pmin)
				*pmin = temp;
	}

	/* check to see if the maximum equals the minimum -- this is a
	   degenerate case -- or it might mean that the input file is
	   faulty. */

	if(*pmax == *pmin)
        {
	  printf ("The input file, %s, is confusing!\n\n", userinputfile);

	  printf ("Either all the points in %s have the same ", userinputfile);
	  printf ("coordinates,\n");
	  printf ("(THE FRACTAL DIMENSION IS ZERO IN THIS CASE.)\n\n");

	  printf ("or %s is simply of the wrong form for an\n",userinputfile);
	  printf ("input file to this program -- please check.\n");

	  exit(-1);
	}

        /* seek to the beginning of the input file so the main program can
	   start there to read in the data. */
        
	/* fseek(infile,0,0); */

	/*  close the input file */
	fclose(infile);
}


void radixsort(array,d1,d2)
	ulong **array;		/* array to be sorted */
	int	d1,d2;		/* dimensions of array */
{
	int  k;			/* control variable */
	Queue      Qzero,Qone; 	/* Queues zero and one. */
	X_Y_marker data;	/* used as data part of queue nodes */

	/* create queues */
	create_Q(&Qone);
	create_Q(&Qzero);

	/* store array of data pairs in Qzero */
	data.marker = false;
	for(k=0;k<d2;k++)
	{
		data.xcoord = array[0][k];
		data.ycoord = array[1][k];
		en_Q(Qzero,data);
	}

	/* set end-of-queue flags */
	data.marker = true;
	en_Q(Qzero,data);
	en_Q(Qone,data);

	/* radix sort queues */
	for(k=0;k<numbits;k++)
	{
		/* sort kth bit of ycoord */
		do
		{
			de_Q(Qzero,&data);
			if(!data.marker)
			{
				if(IS_A_ONE(data.ycoord,k))
					en_Q(Qone,data);
				else
					en_Q(Qzero,data);
			}
		}
		while(!data.marker);

		do
		{
			de_Q(Qone,&data);
			if(!data.marker)
			{
				if(IS_A_ONE(data.ycoord,k))
					en_Q(Qone,data);
				else
					en_Q(Qzero,data);
			}
		}
		while(!data.marker);

		/* set end-of-queue flags */
		data.marker = true;
		en_Q(Qone,data);
		en_Q(Qzero,data);

		/* sort kth bit of xcoord */
		do
		{
			de_Q(Qzero,&data);
			if(!data.marker)
			{
				if(IS_A_ONE(data.xcoord,k))
					en_Q(Qone,data);
				else
					en_Q(Qzero,data);
			}
		}
		while(!data.marker);

		do
		{
			de_Q(Qone,&data);
			if(!data.marker)
			{
				if(IS_A_ONE(data.xcoord,k))
					en_Q(Qone,data);
				else
					en_Q(Qzero,data);
			}
		}
		while(!data.marker);

		/* set end-of-queue flags */
		data.marker = true;
		en_Q(Qone,data);
		en_Q(Qzero,data);
	}

	/* transfer data from Qzero to array */
	k=0;
	do
	{
		de_Q(Qzero,&data);
		if(!data.marker)
		{
			array[0][k] = (ulong)data.xcoord;
			array[1][k++] = (ulong)data.ycoord;
		}
	}
	while(!data.marker);

	/* transfer data from Qone to array */
	do
	{
		de_Q(Qone,&data);
		if(!data.marker)
		{
			array[0][k] = (ulong)data.xcoord;
			array[1][k++] = (ulong)data.ycoord;
		}
	}
	while(!data.marker);

	/* destroy the queues */
        /* ############################## */
       /* This seemed to be generating an error on one of the
	  compilers we tried the program on.  Looks like the problem
	  was something in the destroy_Q function that we fixed. */
	destroy_Q(Qone); 
	destroy_Q(Qzero); 
}


void sweep(array,count_array,d1,d2)
	ulong	**array,		/* array to be sweeped */
		count_array[numbits+1];	/* counter array */
	int	d1,d2;			/* dimensions of array */

{
	ulong	xtest,			 /* XOR of two x values */
		ytest;			 /* XOR of two y values */
	int	i,j,k,m,		 /* control variables */
		flag;			 /* flag to signal bit has changed */

	/* initialize count_array */
	for(i=0;i<=numbits;i++)
		count_array[i]=1;

	/* sweep */
	for(j=0,k=1;k<d2;j++,k++)
	{	/* XOR two x values and two y values */
		xtest=array[0][j]^array[0][k];
		ytest=array[1][j]^array[1][k];

		/* box count */
		i=numbits-1;
		flag=false;
		while (flag == false && i >= 0)
		{	/* check for change in ith bit of x values */
			if(IS_A_ONE(xtest,i))
			{
				for(m=0;m<=i;m++)
					count_array[m]++;
				flag=true;
			}
			else
			{
				/* check for change in ith bit of y values */
				if(IS_A_ONE(ytest,i))
				{
					for(m=0;m<=i;m++)
						count_array[m]++;
					flag=true;
				}
			}
			i--;
		}
	}
}



float fracdim(box_count_array,numpairs)
	ulong	box_count_array[numbits+1],
		numpairs;
{
	int	i,j=0,			/* control variables */
		mark=0;			/* marks the index for > N/5 */
	float	tempfd;			/* The Fractal Dimension to be returned */
	double	xmean=0,ymean=0,	/* mean value for x's and y's */
		temp1=0,temp2=0,	/* variables */
		fdpoints[2][numbits-1];

	/* mark greatest index where count > numpairs/5.  Counts at lesser
	   indexes will not be used in the estimate of fractal dimension --
	   distortion due to smallness of sample is the concern. */
	for(i=0;i<numbits;i++)
	{
	   if(box_count_array[i] > box_count_array[0]/5)
	     mark=i;
	}
        if (mark <= numbits -2)
        {
        printf ("\nThe smallest box size usable in the ");
	printf ("estimate has index:  %d\n", mark);
        printf ("The largest box size used in the estimate ");
	printf ("is at index: %d\n\n", numbits-2) ;
        printf ("Data whose slope will be used to calculate ");
	printf ("Fractal Dimension:\n\n");
	printf("   log[1/e]         log[N(e)]        log[N(e)-N(2e)]\n\n");
        }

	for(i=numbits-2;i>=mark;i--)  
	{
		/* set i'th "x" value to -i -- this is the log (base 2) of
    		   the inverse of the diameter of the i'th size of box.  */
		fdpoints[0][i] = -i;

		/* print the value of x */
                printf ("%8d",-i);

                /* update calculation of the mean x-value. */
		xmean = xmean - i;

		/* set the i'th "y" value to log (base 2) of number    */
		/* of boxes of side 2^i required to cover the set */
		fdpoints[1][i] = log((double) (box_count_array[i]))/log(2.0);

		/* print the value of y */
                printf ("          %12f",fdpoints[1][i]);

		/* print the value of y(i)-y(i-1) */
                if (i<numbits-2)
                printf ("       %12f",fdpoints[1][i]-fdpoints[1][i+1]);
                printf ("\n");

                /* update calculation of the mean y-value. */
	ymean = ymean + log((double) (box_count_array[i]))/log(2.0);
		j++;
	}
	xmean = xmean/((double) j);
	ymean = ymean/((double) j);

	for(i=numbits-2;i>=mark;i--)
	{
	       temp1 = temp1 + ((fdpoints[0][i]-xmean)*(fdpoints[1][i]-ymean));
	       temp2 = temp2 + ((fdpoints[0][i]-xmean)*(fdpoints[0][i]-xmean));
	}

	if(temp2 == 0)
	{
	   printf("\nInsufficient data. More points are needed.\n");
           printf("Cannot assign a Fractal Dimension to this set.\n");
           printf("Examine the data above to make your own conjecture.\n");
/*         tempfd = (float) 0 ; */
           exit(-1) ;   
/*         return(tempfd); */
	}
	else
	{	/* return the Fractal Dimension */
		tempfd = (float) (temp1/temp2);
		return(tempfd);
	}
}
